"""
Used by controller.py to prepare datasets for an experiment, thereby most importantly considering whether the experiment
uses cross validation or not.

Currently, the class is partially hard coded for the poltsanews dataset.
"""
import math
import os
import random
from collections import Counter
from shutil import copyfile

import jsonlines
from tabulate import tabulate

from NewsSentiment.SentimentClasses import SentimentClasses
from NewsSentiment.fxlogger import get_logger


class DatasetPreparer:
    def __init__(self, name, basepath_datasets, task_format, human, non_human):
        # only used for newsmtsc format
        SentimentClasses.Sentiment3ForNewsMtsc()
        self.basepath_datasets = basepath_datasets
        self.name = name
        self.human_created_filenames = human
        self.non_human_created_filenames = non_human
        self.human_created_filepaths = [
            self.get_filepath_by_name(x) for x in self.human_created_filenames
        ]
        self.non_human_created_filepaths = [
            self.get_filepath_by_name(x) for x in self.non_human_created_filenames
        ]
        self.data_types = ["human", "nonhum"]

        self.sets_info = None

        self.random_seed = 1337
        random.seed(self.random_seed)
        self.logger = get_logger()

        self.examples_human = self.files_to_dictlst(self.human_created_filepaths)
        self.examples_nonhum = self.files_to_dictlst(self.non_human_created_filepaths)

        self.task_format = task_format

        self.logger.info(
            "shuffling example lists with seed {}".format(self.random_seed)
        )
        random.shuffle(self.examples_human)
        random.shuffle(self.examples_nonhum)

        self.logger.info(
            "{} examples read created by humans (from: {})".format(
                len(self.examples_human), self.human_created_filepaths
            )
        )
        self.logger.info(
            "{} examples read not created by humans (from: {})".format(
                len(self.examples_nonhum), self.non_human_created_filepaths
            )
        )

    def get_filepath_by_name(self, filename):
        return os.path.join(self.basepath_datasets, self.name, filename)

    def file_to_dictlst(self, filepath):
        dict_lst = []
        with jsonlines.open(filepath, "r") as reader:
            for line in reader:
                dict_lst.append(line)
        self.logger.debug("{} examples read from {}".format(len(dict_lst), filepath))
        return dict_lst

    def files_to_dictlst(self, list_filepaths):
        dict_lst = []
        for filepath in list_filepaths:
            dict_lst.extend(self.file_to_dictlst(filepath))
        return dict_lst

    def create_slices(self, data_type):
        """

        :param sets_info:
        :param data_type: 'human', 'nonhum'
        :return:
        """
        set_names = list(self.sets_info.keys())
        assert data_type in self.data_types

        # get some vars
        _id_relative_weight = "{}-rel-weight".format(data_type)
        _id_examples = "{}-examples".format(data_type)

        prev_split_pos = 0
        for set_index, set_name in enumerate(set_names):
            cur_setinfo = self.sets_info[set_name]

            if _id_relative_weight in cur_setinfo:
                cur_relative_weight = cur_setinfo[_id_relative_weight]

                if cur_relative_weight:
                    split_pos = prev_split_pos + math.floor(
                        cur_relative_weight * len(self.examples_human)
                    )
                    if set_index == len(set_names) - 1:
                        # just to be sure to not miss a single example because of rounding
                        split_pos = len(self.examples_human)

                    example_slice = self.examples_human[prev_split_pos:split_pos]
                    cur_setinfo[_id_examples] = example_slice
                    self.logger.info(
                        "{}: added {} {} examples ({:.2f})".format(
                            set_name, len(example_slice), data_type, cur_relative_weight
                        )
                    )

                    prev_split_pos = split_pos
                else:
                    cur_setinfo[_id_examples] = []

    def merge_slices(self):
        total_examples_count = 0
        for set_name, cur_set in self.sets_info.items():
            cur_merged_data = []
            for data_type in self.data_types:
                _id_examples = "{}-examples".format(data_type)

                if _id_examples in cur_set:
                    cur_merged_data.extend(cur_set[_id_examples])

            cur_set["examples"] = cur_merged_data
            total_examples_count += len(cur_merged_data)

        for set_name, cur_set in self.sets_info.items():
            cur_set["examples-rel"] = len(cur_set["examples"]) / total_examples_count

    def _get_label_counts(self, tasks):
        label_counts = Counter()
        if self.task_format == "newsmtsc":
            for task in tasks:
                assert len(task["targets"]) >= 1, f"no target in task: {task}"
                for target in task["targets"]:
                    label = SentimentClasses.polarity2label(target["polarity"])
                    label_counts[label] += 1
        elif self.task_format == "newstsc":
            for task in tasks:
                label_counts[task["label"]] += 1
        else:
            raise NotImplementedError(f"unknown data format. {self.task_format}")
        return dict(label_counts)

    def print_set_info(self):
        header = [
            "set_name",
            "human rel",
            "human abs",
            "non-hum rel",
            "non-hum abs",
            "rel",
            "abs",
            "pos",
            "neu",
            "neg",
        ]
        rows = []
        for set_name, cur_set in self.sets_info.items():
            if "file" in cur_set:
                num_lines = sum(
                    1 for line in open(self.get_filepath_by_name(cur_set["file"]), encoding="utf8")
                )
                row = [set_name, -1, -1, -1, -1, cur_set["file"], num_lines, -1, -1, -1]
            else:
                label_counts = self._get_label_counts(
                    [*(cur_set["human-examples"]), *(cur_set["nonhum-examples"])]
                )
                row = [
                    set_name,
                    cur_set["human-rel-weight"],
                    len(cur_set["human-examples"]),
                    cur_set["nonhum-rel-weight"],
                    len(cur_set["nonhum-examples"]),
                    cur_set["examples-rel"],
                    len(cur_set["examples"]),
                    label_counts["positive"],
                    label_counts["neutral"],
                    label_counts["negative"],
                ]
            rows.append(row)

        self.logger.info("\n" + tabulate(rows, header))

    def init_set(self, sets_info):
        self.sets_info = sets_info

        set_names = list(self.sets_info.keys())

        weights_human = Counter()
        weights_nonhum = Counter()
        nonnull_set_names = []
        self.filecopy_sets = {}

        # sum weights and thereby filter datasets that are 0 in size
        for set_name in set_names:
            setinfo = self.sets_info[set_name]

            if "file" in setinfo:
                self.filecopy_sets[set_name] = setinfo["file"]
            else:
                weights_human[set_name] = setinfo["human-weight"]
                weights_nonhum[set_name] = setinfo["nonhum-weight"]

                if setinfo["human-weight"] + setinfo["nonhum-weight"] > 0:
                    nonnull_set_names.append(set_name)
                else:
                    self.logger.info(
                        "discard {}, because would be empty".format(set_name)
                    )
        set_names = nonnull_set_names

        human_weight_sum = sum(weights_human.values())
        nonhuman_weight_sum = sum(weights_nonhum.values())

        # add relative weights
        for set_name in set_names:
            cur_setinfo = self.sets_info[set_name]
            if cur_setinfo["human-weight"]:
                cur_setinfo["human-rel-weight"] = (
                    cur_setinfo["human-weight"] / human_weight_sum
                )
            else:
                cur_setinfo["human-rel-weight"] = 0
            if cur_setinfo["nonhum-weight"]:
                cur_setinfo["nonhum-rel-weight"] = (
                    cur_setinfo["nonhum-weight"] / nonhuman_weight_sum
                )
            else:
                cur_setinfo["nonhum-rel-weight"] = 0
        # at this point, setsinfo contains only positive (non-0) relative weights (or no such key, if the abs. pad_value was 0, too)

        # create slices
        self.create_slices("human")
        self.create_slices("nonhum")

        # merge human and non human sets in each set
        self.merge_slices()

        self.print_set_info()

    def export(self, savepath):
        for set_name, cur_set in self.sets_info.items():
            if "file" not in cur_set:
                set_savefolder = os.path.join(savepath, self.name)
                os.makedirs(set_savefolder, exist_ok=True)
                set_savepath = os.path.join(set_savefolder, set_name + ".jsonl")

                with jsonlines.open(set_savepath, "w") as writer:
                    writer.write_all(cur_set["examples"])
                self.logger.debug(
                    "created set (abs={}) at {}".format(
                        len(cur_set["examples"]), set_savepath
                    )
                )

        for set_name, filename in self.filecopy_sets.items():
            set_savefolder = os.path.join(savepath, self.name)
            os.makedirs(set_savefolder, exist_ok=True)
            set_savepath = os.path.join(set_savefolder, set_name + ".jsonl")
            set_sourcepath = self.get_filepath_by_name(filename)

            copyfile(set_sourcepath, set_savepath)
            self.logger.debug("copied set to {}".format(set_savepath))

    @classmethod
    def poltsanews_rel801010_allhuman(cls, basepath):
        name = "poltsanews"
        task_format = "newstsc_old"
        human_created_filenames = ["human.jsonl"]
        non_human_created_filenames = ["train_20191021_233454.jsonl"]

        dprep = cls(
            name,
            basepath,
            task_format,
            human_created_filenames,
            non_human_created_filenames,
        )

        sets_info = {
            "train": {"human-weight": 80, "nonhum-weight": 0},
            "dev": {"human-weight": 10, "nonhum-weight": 0},
            "test": {"human-weight": 10, "nonhum-weight": 0},
        }

        dprep.init_set(sets_info)
        return dprep, name, task_format

    @classmethod
    def poltsanews_crossval8010_allhuman(cls, basepath):
        name = "poltsanews"
        task_format = "newstsc_old"
        human_created_filenames = ["human.jsonl"]
        non_human_created_filenames = ["train_20191021_233454.jsonl"]

        dprep = cls(
            name,
            basepath,
            task_format,
            human_created_filenames,
            non_human_created_filenames,
        )

        sets_info = {
            "crossval": {"human-weight": 80, "nonhum-weight": 0},
            "test": {"human-weight": 10, "nonhum-weight": 0},
        }

        dprep.init_set(sets_info)
        return dprep, name, task_format

    @classmethod
    def acl14twitter(cls, basepath):
        name = "acl14twitter"
        task_format = "newsmtsc"
        human_created_filenames = ["train.raw.jsonl"]
        non_human_created_filenames = []

        dprep = cls(
            name,
            basepath,
            task_format,
            human_created_filenames,
            non_human_created_filenames,
        )

        sets_info = {
            "train": {"human-weight": 80, "nonhum-weight": 0},
            "dev": {"human-weight": 10, "nonhum-weight": 0},
            "test": {"file": "test.raw.jsonl"},
        }

        dprep.init_set(sets_info)
        return dprep, name, task_format

    @classmethod
    def semeval14laptops(cls, basepath):
        name = "semeval14laptops"
        task_format = "newsmtsc"
        human_created_filenames = ["Laptops_Train.xml.seg.jsonl"]
        non_human_created_filenames = []

        dprep = cls(
            name,
            basepath,
            task_format,
            human_created_filenames,
            non_human_created_filenames,
        )

        sets_info = {
            "train": {"human-weight": 80, "nonhum-weight": 0},
            "dev": {"human-weight": 10, "nonhum-weight": 0},
            "test": {"file": "Laptops_Test_Gold.xml.seg.jsonl"},
        }

        dprep.init_set(sets_info)
        return dprep, name, task_format

    @classmethod
    def semeval14restaurants(cls, basepath):
        name = "semeval14restaurants"
        task_format = "newsmtsc"
        human_created_filenames = ["Restaurants_Train.xml.seg.jsonl"]
        non_human_created_filenames = []

        dprep = cls(
            name,
            basepath,
            task_format,
            human_created_filenames,
            non_human_created_filenames,
        )

        sets_info = {
            "train": {"human-weight": 80, "nonhum-weight": 0},
            "dev": {"human-weight": 10, "nonhum-weight": 0},
            "test": {"file": "Restaurants_Test_Gold.xml.seg.jsonl"},
        }

        dprep.init_set(sets_info)
        return dprep, name, task_format

    @classmethod
    def newsmtsc_devtest_mt(cls, basepath):
        name = "newsmtsc-train-and-test-mt"
        task_format = "newsmtsc"
        human_created_filenames = ["devtest_mtsc_only.jsonl"]
        non_human_created_filenames = []
        dprep = cls(
            name,
            basepath,
            task_format,
            human_created_filenames,
            non_human_created_filenames,
        )

        sets_info = {
            "train": {"file": "train.jsonl"},
            "dev": {"human-weight": 30, "nonhum-weight": 0},
            "test": {"human-weight": 70, "nonhum-weight": 0},
        }

        dprep.init_set(sets_info)
        return dprep, name, task_format

    @classmethod
    def newsmtsc_devtest_rw(cls, basepath):
        name = "newsmtsc-train-and-test-rw"
        task_format = "newsmtsc"
        human_created_filenames = ["devtest_mtsc_and_single_primaries.jsonl"]
        non_human_created_filenames = []
        dprep = cls(
            name,
            basepath,
            task_format,
            human_created_filenames,
            non_human_created_filenames,
        )

        sets_info = {
            "train": {"file": "train.jsonl"},
            "dev": {"human-weight": 30, "nonhum-weight": 0},
            "test": {"human-weight": 70, "nonhum-weight": 0},
        }

        dprep.init_set(sets_info)
        return dprep, name, task_format


if __name__ == "__main__":
    DatasetPreparer.sentinews("controller_data/datasets")



# taken from https://github.com/Bjarten/early-stopping-pytorch
# Copyright by Bjarten
# License: MIT

from NewsSentiment.fxlogger import get_logger


class EarlyStopping:
    """Early stops the training if validation loss doesn't improve after a given patience."""

    def __init__(self, patience=2, delta=0.01):
        """
        Args:
            patience (int): How long to wait after last time validation loss improved.
                            Default: 2
            verbose (bool): If True, prints a message for each validation loss improvement.
                            Default: False
            delta (float): Minimum change in the monitored quantity to qualify as an improvement.
                            Default: 0
        """
        self.patience = patience
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.delta = delta
        self.logger = get_logger()
        self.flag_has_score_increased_since_last_check = False

    def __call__(self, dev_score):
        if self.best_score is None:
            self.best_score = dev_score
            self.logger.debug(f'computed first dev-set score {dev_score:.6f}).')
            self.flag_has_score_increased_since_last_check = True
        elif dev_score < self.best_score + self.delta:
            self.counter += 1
            self.logger.debug(
                f'patience counter: {self.counter} out of {self.patience} (cur-score: {dev_score}, best-score:'
                f' {self.best_score})')
            self.flag_has_score_increased_since_last_check = False
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = dev_score
            self.counter = 0
            self.logger.debug(f'dev-set score increased ({self.best_score:.6f} --> {dev_score:.6f}).')
            self.flag_has_score_increased_since_last_check = True



import torch
import torch.nn as nn

# from https://pytorchnlp.readthedocs.io/en/latest/_modules/torchnlp/nn/attention.html
class ClassicAttention(nn.Module):
    """ Applies attention mechanism on the `context` using the `query`.

    **Thank you** to IBM for their initial implementation of :class:`Attention`. Here is
    their `License
    <https://github.com/IBM/pytorch-seq2seq/blob/master/LICENSE>`__.

    Args:
        dimensions (int): Dimensionality of the query and context.
        attention_type (str, optional): How to compute the attention score:

            * dot: :math:`score(H_j,q) = H_j^T q`
            * general: :math:`score(H_j, q) = H_j^T W_a q`

    Example:

         >>> attention = Attention(256)
         >>> query = torch.randn(5, 1, 256)
         >>> context = torch.randn(5, 5, 256)
         >>> output, weights = attention(query, context)
         >>> output.size()
         torch.Size([5, 1, 256])
         >>> weights.size()
         torch.Size([5, 1, 5])
    """

    def __init__(self, dimensions, attention_type="general"):
        super(ClassicAttention, self).__init__()

        if attention_type not in ["dot", "general"]:
            raise ValueError("Invalid attention type selected.")

        self.attention_type = attention_type
        if self.attention_type == "general":
            self.linear_in = nn.Linear(dimensions, dimensions, bias=False)

        self.linear_out = nn.Linear(dimensions * 2, dimensions, bias=False)
        self.softmax = nn.Softmax(dim=-1)
        self.tanh = nn.Tanh()

    def forward(self, query, context):
        """
        Args:
            query (:class:`torch.FloatTensor` [batch size, output length, dimensions]): Sequence of
                queries to query the context.
            context (:class:`torch.FloatTensor` [batch size, query length, dimensions]): Data
                overwhich to apply the attention mechanism.

        Returns:
            :class:`tuple` with `output` and `weights`:
            * **output** (:class:`torch.LongTensor` [batch size, output length, dimensions]):
              Tensor containing the attended features.
            * **weights** (:class:`torch.FloatTensor` [batch size, output length, query length]):
              Tensor containing attention weights.
        """
        batch_size, output_len, dimensions = query.size()
        query_len = context.size(1)

        if self.attention_type == "general":
            query = query.reshape(batch_size * output_len, dimensions)
            query = self.linear_in(query)
            query = query.reshape(batch_size, output_len, dimensions)

        # TODO: Include mask on PADDING_INDEX?

        # (batch_size, output_len, dimensions) * (batch_size, query_len, dimensions) ->
        # (batch_size, output_len, query_len)
        attention_scores = torch.bmm(query, context.transpose(1, 2).contiguous())

        # Compute weights across every context sequence
        attention_scores = attention_scores.view(batch_size * output_len, query_len)
        attention_weights = self.softmax(attention_scores)
        attention_weights = attention_weights.view(batch_size, output_len, query_len)

        # (batch_size, output_len, query_len) * (batch_size, query_len, dimensions) ->
        # (batch_size, output_len, dimensions)
        mix = torch.bmm(attention_weights, context)

        # concat -> (batch_size * output_len, 2*dimensions)
        combined = torch.cat((mix, query), dim=2)
        combined = combined.view(batch_size * output_len, 2 * dimensions)

        # Apply linear_out on every 2nd dimension of concat
        # output -> (batch_size, output_len, dimensions)
        output = self.linear_out(combined).view(batch_size, output_len, dimensions)
        output = self.tanh(output)

        return output, attention_weights



import argparse
import json
from collections import defaultdict

from diskdict import DiskDict

import jsonlines
import pandas as pd

from NewsSentiment.fxlogger import get_logger

logger = get_logger()


def rename_flatten(dictionary, key_prefix):
    new_dict = {}

    for k, v in dictionary.items():
        new_k = key_prefix + "-" + k
        new_dict[new_k] = v

    return new_dict


def without_keys(d, keys):
    return {x: d[x] for x in d if x not in keys}


def non_scalar_to_str(d):
    new_d = {}
    for k, v in d.items():
        if type(v) in [list, dict]:
            new_v = json.dumps(v)
        else:
            new_v = v
        new_d[k] = new_v
    return new_d


COL_NAME_TMP_GROUP = "_tmp_named_id_"


def _find_run_ids(completed_tasks: dict):
    """
    First check the maximum number of run ids, then ensures that each experiment has
    that many.
    :param completed_tasks:
    :return:
    """
    run_ids = set()
    for named_id, result in completed_tasks.items():
        vals = without_keys(result, ["details"])
        run_id = vals["run_id"]
        run_ids.add(run_id)
        del vals["run_id"]
        del vals["experiment_named_id"]
        del vals["experiment_id"]
        named_experiment_id_wo_run_id = json.dumps(vals)
        result[COL_NAME_TMP_GROUP] = named_experiment_id_wo_run_id
    num_runs_per_experiment = len(run_ids)
    logger.info("found %s run_ids: %s", num_runs_per_experiment, run_ids)

    # check that each experiment has as many run ids
    named_id2run_ids = defaultdict(list)
    for named_id, result in completed_tasks.items():
        vals = without_keys(result, ["details"])
        named_experiment_id_wo_run_id = result["_tmp_named_id_"]
        run_id = vals["run_id"]
        named_id2run_ids[named_experiment_id_wo_run_id].append(run_id)

    count_too_few_runs = 0
    for named_id, run_ids in named_id2run_ids.items():
        if len(run_ids) != num_runs_per_experiment:
            logger.debug("%s runs for %s", len(run_ids), named_id)
            count_too_few_runs += 1
    if count_too_few_runs == 0:
        logger.info(
            "GOOD: num experiments with too few runs: %s of %s",
            count_too_few_runs,
            len(named_id2run_ids),
        )
    else:
        logger.warning(
            "num experiments with too few runs: %s of %s",
            count_too_few_runs,
            len(named_id2run_ids),
        )

    return num_runs_per_experiment, completed_tasks


def _aggregate_and_mean(df: pd.DataFrame):
    df = df.copy(deep=True)
    col_names_original_order = list(df.columns)
    df_aggr = df.groupby(COL_NAME_TMP_GROUP).mean()

    # from https://stackoverflow.com/a/35401886
    # this creates a df that contains aggregated values (from df_aggr) and also
    # all other columns (non-aggregated)
    aggr_col_names = list(df_aggr.columns)
    df.drop(aggr_col_names, axis=1, inplace=True)
    df.drop_duplicates(subset=COL_NAME_TMP_GROUP, keep="last", inplace=True)
    df = df.merge(
        right=df_aggr, right_index=True, left_on=COL_NAME_TMP_GROUP, how="right"
    )

    # reorder the dataframe to have the established order of columsn
    # taken from: https://stackoverflow.com/a/13148611
    df = df[col_names_original_order]

    # delete temp col
    del df[COL_NAME_TMP_GROUP]

    return df


def _dfs_to_excel(pathname, name2df):
    writer = pd.ExcelWriter(pathname, engine="xlsxwriter")
    for name, df in name2df.items():
        if df is None:
            logger.info("skipping df because empty: %s", name)
            continue

        df.to_excel(writer, sheet_name=name, startrow=0, startcol=0)
    writer.save()


def shelve2xlsx(opt, ignore_graceful_exit_experiments):
    completed_tasks = DiskDict(opt.results_path)
    logger.info(
        "found {} results in file {}".format(len(completed_tasks), opt.results_path)
    )
    # get max run id
    num_runs_per_experiment, completed_tasks = _find_run_ids(completed_tasks)

    flattened_results = {}

    for named_id, result in completed_tasks.items():
        if result["rc"] == 99 and ignore_graceful_exit_experiments:
            logger.info("found graceful exit (99), not adding to file: %s", named_id)
            continue
        elif result["rc"] == 0:
            test_stats = rename_flatten(result["details"]["test_stats"], "test_stats")
            dev_stats = rename_flatten(result["details"]["dev_stats"], "dev_stats")

            flattened_result = {
                **without_keys(result, ["details"]),
                **dev_stats,
                **test_stats,
            }
        else:
            flattened_result = {**without_keys(result, ["details"])}

        scalared_flattened_result = non_scalar_to_str(flattened_result)
        flattened_results[named_id] = scalared_flattened_result

    df = pd.DataFrame(data=flattened_results.values())

    if num_runs_per_experiment >= 2:
        df_aggr = _aggregate_and_mean(df)
    else:
        df_aggr = None
    del df[COL_NAME_TMP_GROUP]

    _dfs_to_excel(opt.results_path + ".xlsx", {"raw": df, "aggr": df_aggr})


def jsonl2xlsx(opt):
    labels = {2: "positive", 1: "neutral", 0: "negative"}

    with jsonlines.open(opt.results_path, "r") as reader:
        lines = []
        for line in reader:
            if line["true_label"] != line["pred_label"]:
                line["true_label"] = labels[line["true_label"]]
                line["pred_label"] = labels[line["pred_label"]]

                lines.append(line)

        df = pd.DataFrame(data=lines)
        df.to_excel(opt.results_path + ".xlsx")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--results_path",
        type=str,
        default="results/mtscall_stance0",
    )
    parser.add_argument("--mode", type=str, default="shelve")
    opt = parser.parse_args()

    if opt.mode == "shelve":
        shelve2xlsx(opt, ignore_graceful_exit_experiments=False)
    elif opt.mode == "jsonl":
        jsonl2xlsx(opt)



from collections import defaultdict, Counter
from pathlib import Path
import csv
import os

from tqdm import tqdm

from NewsSentiment.fxlogger import get_logger
from NewsSentiment.diskdict import DiskDict

POLARITY2INDEX = {
    "positive": 1,
    "negative": 0,
}
THIS_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
PATH_DICT_BING_LIU_OPINION_POLARITY = THIS_DIR / "opinion_polarity.ddict"

logger = get_logger()


def convert_single_txt_to_dict(path: Path):
    terms = set()
    with open(path, "r") as file:
        line_count = 0
        for line in file:
            if line_count > 29:
                terms.add(line.strip())
            line_count += 1
    return terms


def convert_txt_to_dict():
    path_pos = THIS_DIR / "positive-words.txt"
    path_neg = THIS_DIR / "negative-words.txt"

    term2polarity = defaultdict(set)
    polarity_counter = Counter()

    positive_terms = convert_single_txt_to_dict(path_pos)
    negative_terms = convert_single_txt_to_dict(path_neg)
    all_terms = positive_terms.union(negative_terms)

    for term in all_terms:
        if term in positive_terms:
            term2polarity[term].add(POLARITY2INDEX["positive"])
            polarity_counter["positive"] += 1
        if term in negative_terms:
            term2polarity[term].add(POLARITY2INDEX["negative"])
            polarity_counter["negative"] += 1

    logger.info("read %s terms", len(term2polarity))
    logger.info("polarity count:\n%s", polarity_counter.most_common())

    logger.info("saving to %s...", PATH_DICT_BING_LIU_OPINION_POLARITY)
    ddict_emolex = DiskDict(PATH_DICT_BING_LIU_OPINION_POLARITY)
    ddict_emolex.update(term2polarity)
    ddict_emolex.sync_to_disk()
    logger.info("done")


if __name__ == "__main__":
    convert_txt_to_dict()



def build_trie(lexicon):
    """
    Build a character-trie from the plain pattern_string -> categories_list
    mapping provided by `lexicon`.

    Some LIWC patterns end with a `*` to indicate a wildcard match.
    """
    trie = {}
    for pattern, category_names in lexicon.items():
        cursor = trie
        for char in pattern:
            if char == "*":
                cursor["*"] = category_names
                break
            if char not in cursor:
                cursor[char] = {}
            cursor = cursor[char]
        cursor["$"] = category_names
    return trie


def search_trie(trie, token, token_i=0):
    """
    Search the given character-trie for paths that match the `token` string.
    """
    if "*" in trie:
        return trie["*"]
    if "$" in trie and token_i == len(token):
        return trie["$"]
    if token_i < len(token):
        char = token[token_i]
        if char in trie:
            return search_trie(trie[char], token, token_i + 1)
    return []



import torch

from NewsSentiment.knowledge.liwc.liwchelper import load_token_parser

parse, category_names = load_token_parser()

LIWC_CATEGORY2INDEX = {}
for index, category_name in enumerate(category_names):
    LIWC_CATEGORY2INDEX[category_name] = index


def get_num_liwc_categories():
    return len(category_names)


def get_liwc_categories_as_tensor(term: str):
    categories = parse(term)
    categories_of_lowercased = parse(term.lower())

    if len(categories) == 0 and len(categories_of_lowercased) > 0:
        # if we do not have categories of original term, but have them for lowercased term, use the latter
        categories = categories_of_lowercased

    tensor_emotions = torch.zeros(get_num_liwc_categories(), dtype=torch.long)
    for category in categories:
        index = LIWC_CATEGORY2INDEX[category]
        assert index < get_num_liwc_categories()
        tensor_emotions[index] = 1
    return tensor_emotions



from abc import ABC
from argparse import Namespace
from typing import List, Dict
import torch.nn as nn
import torch

from NewsSentiment.consts import *
from NewsSentiment.dataset import FXDataset
from NewsSentiment.models.FXBaseModel import FXBaseModel


class FXEnsemble(FXBaseModel):
    @staticmethod
    def get_language_models():
        return (
            BERT_BASE_UNCASED,
            ROBERTA_BASE,
            XLNET_BASE_CASED,
        )

    @staticmethod
    def get_input_field_ids():
        return [
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS),
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
            (ROBERTA_BASE, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS),
            (ROBERTA_BASE, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
            (XLNET_BASE_CASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS),
            (XLNET_BASE_CASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
        ]

    def __init__(self, transformer_models: Dict, opt: Namespace):
        super().__init__()

        # language models
        self.bert = transformer_models[BERT_BASE_UNCASED]
        self.roberta = transformer_models[ROBERTA_BASE]
        self.xlnet = transformer_models[XLNET_BASE_CASED]
        self.num_models = 3

        # params
        assert (
            self.bert.config.hidden_size
            == self.roberta.config.hidden_size
            == self.xlnet.config.hidden_size
        )
        self.sequence_length = opt.max_seq_len
        self.hidden_size = self.bert.config.hidden_size

        # other neural network components
        self.dropout = nn.Dropout(opt.dropout)
        self.target_dependent_text_combiner = nn.Linear(
            self.hidden_size * self.num_models, opt.polarities_dim
        )

    def _combine_text_out_with_target_mask(
        self, batch_size, text_last_hidden_state, target_mask
    ):
        roberta_target_mask = target_mask.reshape((batch_size, 1, self.sequence_length))
        roberta_target_dependent_text = torch.bmm(
            roberta_target_mask, text_last_hidden_state
        )
        roberta_target_dependent_text = roberta_target_dependent_text.reshape(
            (batch_size, self.hidden_size)
        )
        return roberta_target_dependent_text

    def forward(self, inputs: List):
        # alternatively, we could also use this
        # FXDataset.get_all_inputs_for_model(input, self)
        bert_text_ids = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS
        )
        bert_target_mask = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK
        )
        roberta_text_ids = FXDataset.get_input_by_params(
            inputs, ROBERTA_BASE, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS
        )
        roberta_target_mask = FXDataset.get_input_by_params(
            inputs, ROBERTA_BASE, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK
        )
        xlnet_text_ids = FXDataset.get_input_by_params(
            inputs, XLNET_BASE_CASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS
        )
        xlnet_target_mask = FXDataset.get_input_by_params(
            inputs, XLNET_BASE_CASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK
        )

        # get variables
        batch_size = bert_text_ids.shape[0]

        # dev notes:
        # batch_size = 4, sequence_length = 150, hidden_size = 768

        # bert_text_out returns list with following elements
        # 0: last_hidden_state (batch_size, sequence_length, hidden_size)
        # 1: pooler_output (batch_size, hidden_size)
        # 2: hidden_states (batch_size, sequence_length, hidden_size)
        (
            bert_text_last_hidden_state,
            bert_text_pooler_output,
            bert_text_hidden_states,
        ) = self.bert(bert_text_ids)
        # roberta_text_out returns same output as bert
        (
            roberta_text_last_hidden_state,
            roberta_text_pooler_output,
            roberta_text_hidden_states,
        ) = self.roberta(roberta_text_ids)
        # xlnet_text_out returns list with following elements
        # 0: last_hidden_state (batch_size, sequence_length, hidden_size)
        # does not exist - (1: mems, a list of length config.n_layers)
        # 2: hidden_states (batch_size, sequence_length, hidden_size)
        xlnet_text_last_hidden_state, xlnet_text_hidden_states = self.xlnet(
            xlnet_text_ids
        )

        # incorporate target masks with (for now) last layer's states
        # *_target_dependent_text_out will be of shape (batch_size, hidden_size)
        bert_target_dependent_text_out = self._combine_text_out_with_target_mask(
            batch_size, bert_text_last_hidden_state, bert_target_mask
        )
        roberta_target_dependent_text_out = self._combine_text_out_with_target_mask(
            batch_size, roberta_text_last_hidden_state, roberta_target_mask
        )
        xlnet_target_dependent_text_out = self._combine_text_out_with_target_mask(
            batch_size, xlnet_text_last_hidden_state, xlnet_target_mask
        )

        # cat outputs
        cat_target_dependent_text_out = torch.cat(
            [
                bert_target_dependent_text_out,
                roberta_target_dependent_text_out,
                xlnet_target_dependent_text_out,
            ],
            dim=1,
        )

        # dropout for better learning
        cat_target_dependent_text_out = self.dropout(cat_target_dependent_text_out)

        # combine and get 3 dimensions
        logits = self.target_dependent_text_combiner(cat_target_dependent_text_out)

        return logits
